#define UINT64_C(c) c##ULL
#include <stdint.h>
#include <stdio.h>
extern "C" {
#include <libavformat/avformat.h>
}

#include <vector>
#include <deque>
#include <algorithm>

static const size_t  SAMPLING_RATE=44100;
static const size_t  ASSUMED_PTS_SCALE      = 1000;
static const size_t  ASSUMED_AUDIO_BYTERATE = (SAMPLING_RATE*2*4);

static const int     FRAMERATE             = 60;
//static const int     FRAMERATE             = 30;
static const size_t  MIN_AUDIO_CORRECTION  = ASSUMED_AUDIO_BYTERATE * 5 / 100; // 0.05 seconds

static const size_t  FRAME_SIZE_IN_AUDIO   = ASSUMED_AUDIO_BYTERATE / FRAMERATE;

static const size_t  WORRYING_AUDIO_PROBLEM_PTS = ASSUMED_PTS_SCALE * 1 / 4; // 0.25 seconds
static const size_t  HALF_FRAME_PTS = ASSUMED_PTS_SCALE / FRAMERATE / 2;

// How much, in terms of "pts", to keep in buffer just in case: 0.5 seconds
static const size_t  BUFFER_REMAIN_PTS = ASSUMED_PTS_SCALE * 5 / 10;

static const size_t  SPLIT_WHENCE      = ASSUMED_PTS_SCALE * 3600*40;

static_assert( ASSUMED_AUDIO_BYTERATE % FRAMERATE == 0,
               "Selected framerate does not provide constant-sized audio packets");
static_assert( (ASSUMED_AUDIO_BYTERATE / FRAMERATE) % 4 == 0,
               "Selected audio rate provides non-4ly divisible audio packet sizes");

//static const unsigned INWIDTH = 320, INHEIGHT = 240;
//static const unsigned INWIDTH = 800, INHEIGHT = 480;
//static const unsigned INWIDTH = 640, INHEIGHT = 360;
static const unsigned INWIDTH = 576, INHEIGHT = 320;

class AVIProducer
{
    AVFormatContext* c;
    AVStream*        video; AVCodec* video_codec;
    AVStream*        audio; AVCodec* audio_codec;

    struct video_frame
    {
        uint64_t pts;
        bool keyframe;
        std::vector<uint8_t> data;
    };

public:
    std::deque<uint8_t>     audio_buffer;
    std::deque<video_frame> video_buffer;
    size_t audio_size_total = 0;
    size_t frames_generated = 0;
    size_t begin_pts = 0;
    size_t file_counter = 0;

public:
    AVIProducer()
    {
        OpenFile();
    }

    void OpenFile()
    {
        char Buf[512] = "papers-combined.avi";
        //sprintf(Buf, "/mnt/hog/uqm-practice-totallyuncut-part-%03d.avi", (int) ++file_counter);
        fprintf(stderr, "Opening new output file: %s\n", Buf);

        avformat_alloc_output_context2(&c, nullptr, "avi", Buf);
        video_codec = avcodec_find_encoder_by_name("zmbv");
        audio_codec = avcodec_find_encoder_by_name("pcm_s32le");
        audio = avformat_new_stream(c, audio_codec); // Stream 0
        video = avformat_new_stream(c, video_codec); // Stream 1

        video->codec->time_base.num = 1;
        video->codec->time_base.den = FRAMERATE;
        video->codec->ticks_per_frame = 1;
        video->codec->pix_fmt = PIX_FMT_BGR0;
        video->codec->width  = INWIDTH;
        video->codec->height = INHEIGHT;
        video->codec->sample_aspect_ratio.num = 1;
        video->codec->sample_aspect_ratio.den = 1;
        video->sample_aspect_ratio.num = 1;
        video->sample_aspect_ratio.den = 1;
        audio->codec->sample_rate = SAMPLING_RATE;
        audio->codec->channels    = 2;
        audio->codec->ticks_per_frame = 1;
        audio->codec->frame_size = 1;//(SAMPLING_RATE*2) / FRAMERATE;

        avio_open(&c->pb, c->filename, AVIO_FLAG_WRITE);
        avformat_write_header(c, nullptr);
    }

    void CloseFile()
    {
        av_write_trailer(c);
    }

    ~AVIProducer()
    {
        CheckFlushing(true);
        CloseFile();
    }

    void AddAudio(uint64_t pts, const uint8_t* data, size_t size)
    {
        size_t pts_to_bytes = pts*ASSUMED_AUDIO_BYTERATE / ASSUMED_PTS_SCALE;

        long dummy=0;
        while(pts_to_bytes >= audio_size_total + MIN_AUDIO_CORRECTION)
        {
            // Generate MIN_AUDIO_CORRECTION bytes of dummy audio
            size_t num_copy = std::min(MIN_AUDIO_CORRECTION, audio_buffer.size());
            //size_t num_zero = MIN_AUDIO_CORRECTION;

            std::vector<uint8_t> dummy_audio( MIN_AUDIO_CORRECTION );
            if(num_copy)
                std::copy(audio_buffer.end() - num_copy,
                          audio_buffer.end(),
                          &dummy_audio[0]);
            audio_buffer.insert(audio_buffer.end(), dummy_audio.begin(), dummy_audio.end());
            dummy += dummy_audio.size();
            audio_size_total += dummy_audio.size();
        }
        if(dummy)
            printf("\n%ld bytes of dummy audio added\n", dummy);

        while(pts_to_bytes + MIN_AUDIO_CORRECTION < audio_size_total)
        {
            size_t correction = MIN_AUDIO_CORRECTION;

            size_t discard    = std::min(correction, audio_buffer.size());
            size_t ignore     = std::min(correction-discard, size);

            if(discard)
            {
                printf("\n%ld bytes of buffered audio discarded\n", (long) discard);
                audio_buffer.erase( audio_buffer.end() - discard, audio_buffer.end() );
                audio_size_total -= discard;
            }
            if(ignore)
            {
                printf("\n%ld bytes of audio discarded\n", (long) ignore);
                size         -= ignore;
                data         += ignore;
                pts_to_bytes += ignore;
            }
            if(!discard && !ignore) return;
        }
        if(!size) return;

        /*printf("Audio at pts=%ld corresponds to %ld bytes; %ld bytes so far\n",
            (long)pts,
            (long)pts_to_bytes,
            (long)audio_size_total);*/

        audio_buffer.insert(audio_buffer.end(), data, data+size);
        audio_size_total += size;

        CheckFlushing(false);
    }

    void AddVideo(uint64_t pts, const uint8_t* data, const size_t size, bool keyframe)
    {
        if(keyframe && pts >= begin_pts + SPLIT_WHENCE)
        {
            CheckFlushing(true);
            CloseFile();
            OpenFile();
            begin_pts = pts;
        }

        uint64_t prev_pts = video_buffer.empty() ? pts : video_buffer.back().pts;
        if(pts < prev_pts)
        {
            uint64_t earlier_pts = prev_pts;
            if(video_buffer.size() >= 2)
                earlier_pts = video_buffer[video_buffer.size()-2].pts;
            uint64_t corrected_pts = (earlier_pts + pts) / 2;

            printf("\nGot a video frame that is before an earlier frame (%ld->%ld->%ld).\n",
                (long)earlier_pts, (long)prev_pts, (long)pts);
            printf("Fixing earlier frame stamp as %ld.\n", (long)corrected_pts);
            video_buffer.back().pts = prev_pts = corrected_pts;
        }
        //printf("video(%ld): pts(%ld), prev_pts(%ld)\n", (long)video_buffer.size(), (long)pts, (long)prev_pts);

        video_buffer.push_back( video_frame{ pts, keyframe, std::vector<uint8_t>{data,data+size} } );

        // If we get video frames, but haven't had audio for a long time,
        // generate dummy audio.
        size_t audio_pts = AudioBytesToPTS(audio_size_total);
        if(prev_pts >= audio_pts + WORRYING_AUDIO_PROBLEM_PTS)
        {
            printf("\nWorrying audio problem detected: got video frame at %ld, but audio is at %ld\n",
                (long)prev_pts, (long)audio_pts);

            AddAudio(prev_pts, nullptr, 0);
        }

        CheckFlushing(false);
    }

    void CheckFlushing(bool final)
    {
        for(;;)
        {
            size_t have_pts = frames_generated * ASSUMED_PTS_SCALE / FRAMERATE;
            size_t margin_pts = have_pts;
            if(!final) margin_pts += BUFFER_REMAIN_PTS;

            size_t audio_pts = AudioBytesToPTS(audio_size_total);

            // Check if we have enough audio in buffer that we can process some video samples
            if(audio_pts >= margin_pts
            && audio_buffer.size() >= FRAME_SIZE_IN_AUDIO
            && (final || video_buffer.size() >= 3))
            {
                // Generate audio packet
                AVPacket p;
                long size = FRAME_SIZE_IN_AUDIO;
                av_new_packet(&p, size);
                p.pts = frames_generated - begin_pts;
                p.dts = frames_generated - begin_pts;
                p.stream_index = 0;    // Audio
                p.flags        = AV_PKT_FLAG_KEY;
                std::copy(audio_buffer.begin(), audio_buffer.begin() + size, p.data);
                av_write_frame(c, &p);
                av_free_packet(&p);
                audio_buffer.erase(audio_buffer.begin(), audio_buffer.begin() + size);

                // Generate video packet, if there is one in queue
                if(!video_buffer.empty()
                && video_buffer.front().pts <= have_pts + HALF_FRAME_PTS)
                {
                    //printf("Video out: %ld\n", video_buffer.front().pts);
                    const auto& v = video_buffer.front();
                    av_new_packet(&p, v.data.size());
                    p.pts = frames_generated - begin_pts;
                    p.dts = frames_generated - begin_pts;
                    p.stream_index = 1;    // Video
                    p.flags        = v.keyframe ? AV_PKT_FLAG_KEY : 0;
                    std::copy(v.data.begin(), v.data.end(), p.data);
                    av_write_frame(c, &p);
                    av_free_packet(&p);
                    video_buffer.pop_front();
                }
                else
                {
                    // Generate blank video packet
                    av_new_packet(&p, 0);
                    p.pts = frames_generated - begin_pts;
                    p.dts = frames_generated - begin_pts;
                    p.stream_index = 1;    // Video
                    p.flags        = 0;
                    av_write_frame(c, &p);
                    av_free_packet(&p);
                }
                ++frames_generated;
            }
            else
                break;
        }
    }

    size_t PTStoAudioBytes(uint64_t pts) const
    {
        return pts * ASSUMED_AUDIO_BYTERATE / ASSUMED_PTS_SCALE;
    }
    uint64_t AudioBytesToPTS(size_t bytes) const
    {
        return bytes * ASSUMED_PTS_SCALE / ASSUMED_AUDIO_BYTERATE;
    }

    uint64_t FlushAndGetPTS()
    {
        CheckFlushing(true);
        return frames_generated * ASSUMED_PTS_SCALE / FRAMERATE;
    }
};

int main(int argc, char** argv)
{
    av_register_all();

    AVIProducer produce;

    uint64_t base_pts = 0;
    uint64_t in_audio_pts = 0;
    uint64_t in_video_pts = 0;

    for(int a=1; a<argc; ++a)
    {
        printf("\n\nOpening file %d/%d: %s\n", a,argc, argv[a]);
        fprintf(stderr, "Opening file %d/%d: %s\n", a,argc, argv[a]);

        AVFormatContext* c = nullptr;
        int i = avformat_open_input(&c, argv[a], nullptr, nullptr);
        av_dump_format(c, 0, argv[a], 0);
        if(i < 0) continue;

        int video_id = av_find_best_stream(c, AVMEDIA_TYPE_VIDEO, -1,-1, nullptr, 0);
        int audio_id = av_find_best_stream(c, AVMEDIA_TYPE_AUDIO, -1,-1, nullptr, 0);

        printf("Video stream is %d, audio stream is %d\n", video_id, audio_id);

        AVPacket p;

        for(;;)
        {
            av_init_packet(&p);
            int j = av_read_frame(c, &p);
            if(j < 0) break;

            bool keyframe = p.flags & AV_PKT_FLAG_KEY;
            bool audio    = p.stream_index == audio_id;

#if 1
            printf("\nj(%d) stream(%d) flags(%02X) pts(%ld - %ld) duration(%ld) size(%p,%d)\n",
                j,
                p.stream_index,
                p.flags,
                (long) p.pts + base_pts, (long)base_pts,
                (long) p.duration,
                p.data, p.size);
            fflush(stdout);
#endif
            if(audio)
                produce.AddAudio(in_audio_pts = p.pts + base_pts, p.data, p.size);
            else
                produce.AddVideo(in_video_pts = p.pts + base_pts, p.data, p.size, keyframe);

            printf("\rBase(%ld) V(%ld) A(%ld) Frames(%ld, V:%ld, A:%ld, BufV:%ld, BufA:%ld)",
                (long)base_pts,
                (long)in_video_pts,
                (long)in_audio_pts,
                (long) produce.frames_generated,
                (long) (produce.frames_generated * ASSUMED_PTS_SCALE / FRAMERATE),
                (long) (produce.audio_size_total * ASSUMED_PTS_SCALE / ASSUMED_AUDIO_BYTERATE),
                (long) produce.video_buffer.size(),
                (long) produce.audio_buffer.size() );
            fflush(stdout);

            av_free_packet(&p);
        }
        avformat_close_input(&c);

        base_pts = produce.FlushAndGetPTS();
    }
}
