/*** SETTINGS ***/
#define OUTPUT_FILE_PATTERN "/tmp/xwindow%llu.mkv"
/*** END SETTINGS ***/

#define USE_FORK 1


#include <dlfcn.h>
#include <fcntl.h>
#include <unistd.h>

#include <cstdio>
#include <vector>
#include <algorithm>

#include <sys/time.h>

#include <X11/Xlib.h>
#include <xcb/shm.h>

#include <errno.h>
#include <sys/wait.h>

#include <sys/shm.h>
#include <map>

#define CALL_OLD

#include "zmbv_dosbox.h"
#include "libmkv.h"
#include "rgbtorgb.hh"

/*template<const char* libname, const char* funcname, typename T, typename...P>
struct caller
{
    T (*orig) (P...);
public:
    caller()
    {
        void* f = dlopen(libname, RTLD_NOW);
        if(!f) perror(libname)
        orig = (T(*)(P...)) dlsym(f, funcname);
        dlclose(f);
    }
};*/

static void FullyWrite(int fd, const void* b, unsigned length)
{
    const unsigned char* buf = (const unsigned char*) b;
  Retry:;
    int result = write(fd, buf, length);
    if(result == -1 && errno==EAGAIN) goto Retry;
    if(result == -1 && errno==EINTR) goto Retry;
    if(result <= 0) { fprintf(stderr, "[pid %d:result=%d/%u]",getpid(),result,length);
                      perror("write"); return; }
    length -= result;
    buf    += result;
    if(length) goto Retry;
}
static void FullyRead(int fd, void* b, unsigned length)
{
    unsigned char* buf = (unsigned char*) b;
  Retry:;
    int result = read(fd, buf, length);
    if(result == -1 && errno==EAGAIN) goto Retry;
    if(result == -1 && errno==EINTR) goto Retry;
    if(result <= 0) { fprintf(stderr, "[pid %d:result=%d/%u]",getpid(),result,length);
                      perror("read"); return; }
    length -= result;
    buf    += result;
    if(length) goto Retry;
}

typedef ZMBV_Codec::rect rect;

struct encoder
{
    ZMBV_Codec zmbv;
    size_t     bufsize;
    char*      buffer;
    unsigned   framecount;
    zmbv_format_t format;
    mk_Writer* mkw;
    mk_Track*  mkt_v;
    mk_Track*  mkt_a;
    int64_t    first_time, last_time;
    unsigned   bytesperpixel;
    std::vector<unsigned char> FrameBuffer;
    std::vector<void*        > LinePointers;

    struct BufferedFrame
    {
        bool    audio;

        int64_t time;
        bool    key;
        std::vector<unsigned char> data;
    };
    std::vector<BufferedFrame> buffered;

    std::string current_filename;
    std::string current_sequence;

    pthread_mutex_t mut;

    encoder() : zmbv(), bufsize(0), buffer(0), framecount(0), format(ZMBV_FORMAT_NONE),
        mkw(0),
        mkt_v(0),
        mkt_a(0)
    {
        pthread_mutex_init(&mut, 0);
    }
    ~encoder()
    {
        Close();
        pthread_mutex_destroy(&mut);
    }
    void Close()
    {
        fprintf(stderr, "XPATCH-X11: Closing ZMBV & MKV\n"); fflush(stderr);
        delete[] buffer;
        buffer = 0;
        #pragma omp critical(mkv_access)
        {
        mk_close(mkw);
        }
    }
    void Reset()
    {
        fprintf(stderr, "XPATCH-X11: Beginning new ZMBV\n"); fflush(stderr);

        zmbv.~ZMBV_Codec();
        new(&zmbv) ZMBV_Codec{};

        bufsize = 0;
        buffer = 0;
        framecount = 0;
        format = ZMBV_FORMAT_NONE;
        first_time = 0;
        last_time = 0;
        FrameBuffer.clear();
        LinePointers.clear();
        mkw = 0;
        mkt_v = 0;
        mkt_a = 0;
        buffered.clear();
    }

    void MKW_Open()
    {
        if(mkw) return;

        char Buf[256];
        int64_t cur_time;
        struct timeval tv;
        gettimeofday(&tv, NULL);
        cur_time = (tv.tv_usec + (int64_t)tv.tv_sec * 1000000) * 1000ll;
        sprintf(Buf, OUTPUT_FILE_PATTERN, (long long)cur_time);
        current_filename = Buf;

        #pragma omp critical(mkv_access)
        {
        mkw = mk_createWriter(current_filename.c_str(), 1000000, 1);
        }
    }

    void MKW_Header()
    {
        #pragma omp critical(mkv_access)
        {
        if(mkt_a && mkt_v)
        {
            mk_writeHeader(mkw, "xpatch");
            for(const auto& f: buffered)
            {
                mk_Track* which = f.audio ? mkt_a : mkt_v;
                //fprintf(stderr, "Puts cached %s data\n", f.audio?"audio":"video");
                mk_startFrame(mkw, which);
                mk_addFrameData(mkw, which, &f.data[0], f.data.size() );
                mk_setFrameFlags(mkw, which, f.time, f.key, 0);
                mk_flushFrame(mkw, which);
            }
            buffered.clear();
        }
        }
    }

    void MKW_Put(bool audio, void* stream, unsigned bytes, int64_t time, bool key)
    {
        time -= first_time;

        #pragma omp critical(mkv_access)
        {
        if(mkt_a && mkt_v)
        {
            mk_Track* which = audio ? mkt_a : mkt_v;
            //fprintf(stderr, "Puts %s data\n", audio?"audio":"video");

            mk_startFrame(mkw, which);
            mk_addFrameData(mkw, which, stream, bytes);
            mk_setFrameFlags(mkw, which, time, key, 0);
            mk_flushFrame(mkw, which);
        }
        else
        {
            buffered.push_back( BufferedFrame{ audio, time, key,
                                               std::vector<unsigned char>
                                               ( (const unsigned char*)stream,
                                                 (const unsigned char*)stream + bytes )
                                             } );
        }
        }
    }

    void Blit(
        unsigned image_width, unsigned image_height,
        unsigned image_bits_per_pixel,
        char* image_data,
        unsigned image_bytes_per_line,
        rect& rect,
        int64_t cur_time = 0)
    {
        pthread_mutex_lock(&mut);
        struct unlocker
        {
            pthread_mutex_t* m;
            ~unlocker()
            {
                pthread_mutex_unlock(m);
            }
        } unlocker = { &mut };

        if(image_height != LinePointers.size() && mkt_v)
        {
            Reset();
        }

        if(!mkt_v && !mkt_a) first_time = cur_time;

        if(!mkt_v)
        {
#ifdef CONVERT_TO_16BPP
            format        = ZMBV_FORMAT_16BPP;
            bytesperpixel = 2;
#else
            format        = zmbv.BPPFormat(image_bits_per_pixel);
            bytesperpixel = (image_bits_per_pixel+7)/8;
#endif

            fprintf(stderr, "Image format: %ux%u, %u bpp,%u bytes per pixel,pitch=%u\n",
                image_width,
                image_height,
                image_bits_per_pixel,
                (image_bits_per_pixel+7)/8,
                image_bytes_per_line);
            if(!zmbv.SetupCompress(image_width, image_height))
                fprintf(stderr, "setupcompress failed\n");
            else
                fprintf(stderr, "setupcompress ok\n");

            mk_TrackConfig mkc_v;
            memset(&mkc_v, 0, sizeof(mkc_v));
            mkc_v.trackType  = MK_TRACK_VIDEO;
            mkc_v.flagLacing = MK_LACING_NONE;
            mkc_v.flagEnabled = 1;
            mkc_v.flagDefault = 1;
            mkc_v.defaultDuration = 0;//(1/60.0) * 1e9;
            mkc_v.name = 0;
            mkc_v.codecID   = "V_MS/VFW/FOURCC";
            mkc_v.codecName = "ZMBV";
            char buf[40] = { 0 };
            memcpy(buf+16, CODEC_4CC, 4); // MPlayer looks for fourcc in this offset
            mkc_v.codecPrivate     = buf;
            mkc_v.codecPrivateSize = 40;    // MPlayer requires at least this much

            mkc_v.extra.video.pixelWidth = image_width;
            mkc_v.extra.video.pixelHeight = image_height;
            mkc_v.extra.video.displayWidth  = image_width;
            mkc_v.extra.video.displayHeight = image_height;
            mkc_v.extra.video.aspectRatioType = MK_ASPECTRATIO_FIXED;

            fprintf(stderr, "Opening MKW\n");

            MKW_Open();
            mkt_v = mk_createTrack(mkw, &mkc_v);
            MKW_Header();

            fprintf(stderr, "MKW header generated (maybe)\n");

            last_time = cur_time;

            bufsize = zmbv.NeededSize(image_width, image_height, format);
            buffer  = new char[bufsize];
            fprintf(stderr, "zmbv buffer at %p, %zu bytes\n", buffer, bufsize);

            FrameBuffer.resize( image_width * image_height * bytesperpixel );
            LinePointers.resize( image_height );
            for(unsigned y=0; y< image_height; ++y)
                LinePointers[y] = (void*) &FrameBuffer[y * image_width * bytesperpixel];
        }


        bool do_key = false;
        if(framecount%120 == 0 || (cur_time - last_time) > 1000000000ll)
        {
            do_key=true;
        }

        if(!zmbv.PrepareCompressFrame( do_key+0, format, 0, buffer, bufsize, 8,8))
        {
            fprintf(stderr, "preparecompressframe failed\n");
        }

        unsigned sbpp = (image_bits_per_pixel+7)/8;
        char* p = (char*) image_data;

        for(unsigned yo=0; yo< rect.h; ++yo)
        {
            unsigned y = rect.y + yo;
            unsigned x = rect.x;
            unsigned nx = rect.w;
            char *q = (char*) &p[ y * image_bytes_per_line + x * sbpp ];
            char* t = (char*) &FrameBuffer[ (y * image_width + x) * bytesperpixel ];
#ifdef CONVERT_TO_16BPP
            if(sbpp == 2)
                memcpy(t, q, nx*2);
            else if(sbpp == 4)
                Convert32ToR16Frame(q, (unsigned char*) t, nx, nx, y);
            else
                Convert24ToR16Frame(q, (unsigned char*) t, nx, nx, y);
#else
            memcpy(t, q, nx * bytesperpixel);
#endif
        }

        ZMBV_Codec::rect tmp = {(unsigned)rect.x,(unsigned)rect.y, (unsigned)rect.w,(unsigned)rect.h};
        if(do_key)
            tmp = { 0u, 0u, (unsigned) image_width, (unsigned) image_height};

        zmbv.CompressLines(image_height, &LinePointers[0], tmp);
        size_t gotsize = zmbv.FinishCompressFrame(tmp);

        MKW_Put(false, buffer, gotsize, cur_time, do_key);

        fprintf(stderr, "Frame %u at time=%llu, %u byte pixels, zmbv format %u, key=%s; size = %zu (%ux%u << %ux%u)\n",
            framecount, (unsigned long long) (cur_time - first_time),
            bytesperpixel, format,
            do_key?"true":"false",
            gotsize,  image_width,image_height, rect.w,rect.h);

        ++framecount;
        last_time   = cur_time;
    }

    void AddAudio(unsigned char* stream, unsigned bytes, int64_t cur_time)
    {
        pthread_mutex_lock(&mut);
        struct unlocker
        {
            pthread_mutex_t* m;
            ~unlocker()
            {
                pthread_mutex_unlock(m);
            }
        } unlocker = { &mut };

        if(!mkt_v && !mkt_a) first_time = cur_time;

        if(!mkt_a)
        {
            mk_TrackConfig mkc_a;
            memset(&mkc_a, 0, sizeof(mkc_a));
            mkc_a.trackType  = MK_TRACK_AUDIO;
            mkc_a.flagLacing = MK_LACING_NONE;
            mkc_a.flagEnabled = 1;
            mkc_a.flagDefault = 1;
            mkc_a.defaultDuration = 0;//(1/60.0) * 1e9;
            mkc_a.name = 0;
            mkc_a.codecID  = const_cast<char*>(MK_ACODEC_PCMINTLE);
            //mkc_a.codecID  = const_cast<char*>(MK_ACODEC_PCMFLTLE);
            mkc_a.extra.audio.samplingFreq = 44100;
            mkc_a.extra.audio.channels     = 2;
            mkc_a.extra.audio.bitDepth     = 32;

            MKW_Open();
            mkt_a = mk_createTrack(mkw, &mkc_a);
            MKW_Header();
        }

        //fprintf(stderr, "Dealing with audio: %p, %u bytes\n", stream,bytes);
        MKW_Put(true, stream, bytes, cur_time, true);
    }

    #if 0
    void AppendCurrentSequence(int save_key)
    {
        current_sequence += current_filename;
        current_sequence += '\n';

        char Buf[256];
        sprintf(Buf, SEQUENCE_FILE_PATTERN, save_key);
        FILE* fp = fopen(Buf, "wt");
        fwrite(current_sequence.data(), 1, current_sequence.size(), fp),
        fclose(fp);

        fprintf(stderr, "XPATCH-X11: Saved current sequence as save %d\n", save_key);
    }
    void LoadSequence(int save_key)
    {
        fprintf(stderr, "XPATCH-X11: Discarding current sequence, loads save %d\n", save_key);
        current_sequence.clear();

        char Buf[256];
        sprintf(Buf, SEQUENCE_FILE_PATTERN, save_key);
        FILE* fp = fopen(Buf, "rt+");
        if(fp)
        {
            fseek(fp, 0, SEEK_END);
            long size = ftell(fp);
            rewind(fp);
            std::vector<char> tmp( size );
            fread(&tmp[0], 1, size, fp);
            fclose(fp);
            tmp.push_back('\0');
            current_sequence = &tmp[0]; // Read nul-terminated string
            fprintf(stderr, "XPATCH-X11: Appending to sequence corresponding to save %d\n", save_key);
        }
        else
        {
            perror(Buf);
            fprintf(stderr, "Unknown save sequence!\n");
        }
    }
    #endif
};

struct sender
{
    encoder enc;
    bool    first = true;
    //unsigned   cur_w = ~0u, cur_h = ~0u;
    //unsigned   output_file_counter = 0;

    pthread_mutex_t mut;

    int pid = -1, pip[2];

    sender()
    {
    }
    ~sender()
    {
    #if USE_FORK
        if(pid != 0)
        {
            char a = 'Q';
            FullyWrite(pip[1], &a, 1);
            close(pip[1]);
            waitpid(pid, NULL, 0);
        }
        else
    #endif
            Close();
        pthread_mutex_destroy(&mut);
    }
    void Close()
    {
        enc.Close();
    }
    void Blit(
        unsigned image_width, unsigned image_height,
        unsigned image_bits_per_pixel,
        char* image_data,
        unsigned image_bytes_per_line,
        rect& rect)
    {
        unsigned BytesPerPixel = (image_bits_per_pixel + 7) / 8;
        unsigned BitsPerPixel  = image_bits_per_pixel;

        if(image_width < 64 || image_height < 64)
        {
            fprintf(stderr, "Ignoring video frame of %ux%u\n",
                image_width, image_height);
            return;
        }

        struct timeval tv;
        gettimeofday(&tv, NULL);
        int64_t cur_time = (tv.tv_usec + (int64_t)tv.tv_sec * 1000000) * 1000ll;

    #if USE_FORK
        OpenParent();

        #if 0
        // Parent: Send events
        if(image_width != cur_w
        || image_height != cur_h)
        {
            fprintf(stderr, "Resolution changed from %ux%u to %ux%u, restarting video\n",
                cur_w,cur_h,
                image_width, image_height);
            char v = 'Q';
            FullyWrite(pip[1], &v, 1);
            close(pip[1]);
            waitpid(pid, NULL, 0);
            first = true;
            pid          = -1;
            ++output_file_counter;
            goto reconfigure_parent;
        }
        #endif

       {char v = 'V';
        pthread_mutex_lock(&mut);
        struct unlocker{ pthread_mutex_t* m; ~unlocker() { pthread_mutex_unlock(m); } } unlocker = { &mut };

        FullyWrite(pip[1], &v, 1);
        FullyWrite(pip[1], &cur_time, sizeof(cur_time));
        FullyWrite(pip[1], &rect, sizeof(rect));
        FullyWrite(pip[1], &image_width, sizeof(image_width));
        FullyWrite(pip[1], &image_height, sizeof(image_height));
        FullyWrite(pip[1], &image_bits_per_pixel, sizeof(image_bits_per_pixel));
        FullyWrite(pip[1], &image_bytes_per_line, sizeof(image_bytes_per_line));
        const char* p = image_data;
        unsigned src_pitch = image_bytes_per_line;
        for(unsigned yo=0; yo< rect.h; ++yo)
        {
            unsigned y = rect.y + yo;
            unsigned x = rect.x;
            unsigned nx = rect.w;
            const char *q = (const char*) &p[ y * src_pitch + x * BytesPerPixel ];
            FullyWrite(pip[1], q, nx * BytesPerPixel);
        }}
        return;
    #else
        enc.Blit(image_width,image_height,image_bits_per_pixel,image_data,image_bytes_per_line,rect,cur_time);
    #endif
    }

#ifdef USE_FORK
    void OpenParent()
    {
        pthread_mutex_lock(&mut);
        struct unlocker{ pthread_mutex_t* m; ~unlocker() { pthread_mutex_unlock(m); } } unlocker = { &mut };

        if(first)
        {
        //reconfigure_parent:;
            fprintf(stderr, "xpatch-parent pid: %d\n", getpid());
            // xpatch -> pip2 -> bufferprog -> pipe -> mkv&zmbv
            int pip2[2];
            pipe(pip2);
            pid = fork();
            if(pid)
            {
                close(pip2[0]);
                pip[1] = pip2[1];
                first = false;
                //cur_w = image_width;
                //cur_h = image_height;
                return; // parent
            }
            // Close everything except the pipe and stderr
            {
              #ifndef OPEN_MAX
               #define OPEN_MAX 1024
              #endif
              int q = getdtablesize(); if(q < 0) q = OPEN_MAX;
              for(int p=0; p<q; ++p)
                  if(p != 2 && p != pip2[0])
                      close(p);
            }
            pipe(pip);
            if(!fork())
            {
                // Rename pip2[0] into 0, pip[1] into 1.
                close(pip[0]);
                if(pip[1] == 0) pip[1] = dup(pip[1]); // Ensure stdout is not 0
                if(pip2[0] != 0) { dup2(pip2[0], 0); close(pip2[0]); }
                if(pip[1]  != 1) { dup2(pip[1],  1); close(pip[1]);  }

                if(nice(-20) < 0) // set maximum priority for bufferprog
                    perror("nice"); // doesn't really matter if it fails.

                unsetenv("LD_PRELOAD");
                execlp("bufferprog", "bufferprog", "500e6", NULL);
                perror("execlp"); fflush(stderr);
                _exit(-1);
            }
            nice(10); // set minimum priority to child
            close(pip2[0]);
            close(pip[1]);
            fprintf(stderr, "xpatch-child pid: %d\n", getpid()); fflush(stderr);

        //child:;
            // Child. Wait for events
            bool terminate = false;
            while(!terminate)
            {
                char c = 'Q';
                int64_t cur_time = 0;
                FullyRead(pip[0], &c, 1);
                switch(c)
                {
                    case 'V':
                    {
                        unsigned image_width, image_height, image_bits_per_pixel, image_bytes_per_line;
                        rect rect;

                        FullyRead(pip[0], &cur_time,     sizeof(cur_time));
                        FullyRead(pip[0], &rect,         sizeof(rect));
                        FullyRead(pip[0], &image_width,  sizeof(image_width));
                        FullyRead(pip[0], &image_height, sizeof(image_height));
                        FullyRead(pip[0], &image_bits_per_pixel, sizeof(image_bits_per_pixel));
                        FullyRead(pip[0], &image_bytes_per_line, sizeof(image_bytes_per_line));
                        unsigned BytesPerPixel = (image_bits_per_pixel + 7) / 8;
                        //fprintf(stderr, "V size: %ux%ux%u\n", rect.w,rect.h,BytesPerPixel);
                        std::vector<char> buffer( rect.w * rect.h * BytesPerPixel);
                        FullyRead(pip[0], &buffer[0], buffer.size());
                        enc.Blit(image_width,image_height,image_bits_per_pixel,&buffer[0],image_bytes_per_line,rect,cur_time);
                        break;
                    }
                    case 'A':
                    {
                        unsigned bytes = 0;
                        FullyRead(pip[0], &cur_time,     sizeof(cur_time));
                        FullyRead(pip[0], &bytes,        sizeof(bytes));
                        //fprintf(stderr, "A size: %u\n", bytes);
                        std::vector<unsigned char> buffer(bytes);
                        FullyRead(pip[0], &buffer[0], buffer.size());
                        enc.AddAudio(&buffer[0], buffer.size(), cur_time);
                        break;
                    }
                    case 'Q':
                        terminate = true;
                        break;
                }
            }
            Close();
            _exit(0);
        }
    }
#endif

    void AddAudio(unsigned char* stream, unsigned bytes)
    {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        int64_t cur_time = (tv.tv_usec + (int64_t)tv.tv_sec * 1000000) * 1000ll;

#ifdef USE_FORK
        OpenParent();

       {char v = 'A';
        pthread_mutex_lock(&mut);
        struct unlocker{ pthread_mutex_t* m; ~unlocker() { pthread_mutex_unlock(m); } } unlocker = { &mut };

        FullyWrite(pip[1], &v, 1);
        FullyWrite(pip[1], &cur_time, sizeof(cur_time));
        FullyWrite(pip[1], &bytes, sizeof(bytes));
        FullyWrite(pip[1], stream, bytes);
       }
        return;
#else
        enc.AddAudio(stream, bytes, cur_time);
#endif
    }

    void Blit(
        XImage* image,
        rect& rect)
    {
        Blit(image->width,image->height,
             image->bits_per_pixel,
             image->data,
             image->bytes_per_line,
             rect);
    }
};

static std::map<int/*shmid*/,const void*> shmid_mappings;
static std::map<xcb_shm_seg_t,int>        shmseg_mappings;
static int dsp_filehandle = -1;

static sender& GetSender()
{
    static sender s;
    return s;
}

extern "C" {


#define XPutImage_f_params \
    Display*         display , \
    Drawable         d , \
    GC           gc ,    \
    XImage*      image , \
    int          src_x , \
    int          src_y , \
    int          dest_x , \
    int          dest_y , \
    unsigned int     width , \
    unsigned int     height 

int XPutImage(XPutImage_f_params)
{
    //fprintf(stderr, "XPutImage detected\n");
    /* This object will load the _real_ XPutImage
     * symbol and enable calling it
     */
    static struct caller
    {
        int (*orig) (XPutImage_f_params);
    public:
        caller()
        {
            void *f = dlopen("libXext.so", RTLD_NOW);
            if(!f) perror("libXext.so");
            orig = (int(*)(XPutImage_f_params))dlsym(f, "XPutImage");
            dlclose(f);
        }
    } caller;
    auto& sender = GetSender();
    //if(!image) { sender.Close(); return 0; }
    int res = 0;
#ifdef CALL_OLD
    res = caller.orig(display,d,gc,image,src_x,src_y,dest_x,dest_y,width,height);
#endif
    /*fprintf(stderr, "Blit im(%u,%u,depth=%u,pitch=%u,bpp=%u) src(%u,%u) dest(%u,%u) size(%u,%u)\n",
        image->width, image->height,
        image->depth, image->bytes_per_line,
        image->bits_per_pixel,
        src_x,src_y, dest_x,dest_y, width,height); fflush(stderr);*/
    rect r = { src_x,src_y, width,height };
    sender.Blit(image, r);
    return res;
}



#define XShmPutImage_f_params \
    Display*         dpy , \
    Drawable         d , \
    GC           gc , \
    XImage*      image , \
    int          src_x ,\
    int          src_y , \
    int          dst_x , \
    int          dst_y ,  \
    unsigned int     src_width , \
    unsigned int     src_height , \
    Bool         send_event 

int XShmPutImage(XShmPutImage_f_params)
{
    fprintf(stderr, "XShmPutImage detected\n");
    static struct caller
    {
        int (*orig) (XShmPutImage_f_params);
    public:
        caller()
        {
            void *f = dlopen("libXext.so", RTLD_NOW);
            if(!f) perror("libXext.so");
            orig = (int(*)(XShmPutImage_f_params))dlsym(f, "XShmPutImage");
            dlclose(f);
        }
    } caller;
    auto& sender = GetSender();
    //if(!image) { sender.Close(); return 0; }
    int res = 0;
#ifdef CALL_OLD
    res = caller.orig(dpy,d,gc,image,src_x,src_y,dst_x,dst_y,src_width,src_height,send_event);
#endif
    /*fprintf(stderr, "Blit im(%u,%u,depth=%u,pitch=%u,bpp=%u) src(%u,%u) dest(%u,%u) size(%u,%u)\n",
        image->width, image->height,
        image->depth, image->bytes_per_line,
        image->bits_per_pixel,
        src_x,src_y, dest_x,dest_y, width,height);*/
    rect r = { src_x,src_y, src_width,src_height };
    sender.Blit(image, r);
    return res;
}


#define xcb_shm_put_image_params \
                   xcb_connection_t *c  /**< */,\
                   xcb_drawable_t    drawable  /**< */,  \
                   xcb_gcontext_t    gc  /**< */,\
                   uint16_t          total_width  /**< */,\
                   uint16_t          total_height  /**< */,\
                   uint16_t          src_x  /**< */,\
                   uint16_t          src_y  /**< */,\
                   uint16_t          src_width  /**< */,\
                   uint16_t          src_height  /**< */,\
                   int16_t           dst_x  /**< */,\
                   int16_t           dst_y  /**< */,\
                   uint8_t           depth  /**< */,\
                   uint8_t           format  /**< */,\
                   uint8_t           send_event  /**< */,\
                   xcb_shm_seg_t     shmseg  /**< */,\
                   uint32_t          offset  /**< */\

xcb_void_cookie_t xcb_shm_put_image(xcb_shm_put_image_params)
{
    fprintf(stderr, "xcb_shm_put_image detected\n");
    static struct caller
    {
        xcb_void_cookie_t (*orig) (xcb_shm_put_image_params);
    public:
        caller()
        {
            void *f = dlopen("libxcb-shm.so", RTLD_NOW);
            if(!f) perror("libxcb-shm.so");
            orig = (xcb_void_cookie_t(*)(xcb_shm_put_image_params))dlsym(f, "xcb_shm_put_image");
            if(!orig) perror("dlsym");
            dlclose(f);
        }
    } caller;
    auto& sender = GetSender();
    //if(!image) { sender.Close(); return 0; }
    xcb_void_cookie_t res;

    const void* ptr = nullptr;
    auto i = shmseg_mappings.find(shmseg);
    if(i != shmseg_mappings.end())
    {
        auto j = shmid_mappings.find(i->second);
        if(j != shmid_mappings.end())
            ptr = j->second;
    }

    /*fprintf(stderr, "SHM put image tot=%ux%u src=%u,%u %ux%u dst=%u,%u depth=%u offset=%u at=0x%X/%p\n",
        total_width,total_height, src_x,src_y,src_width,src_height, dst_x,dst_y, depth, offset, shmseg,ptr);*/

#ifdef CALL_OLD
    res = caller.orig(c,drawable,gc,total_width,total_height,src_x,src_y,src_width,src_height,dst_x,dst_y,depth,format,send_event,shmseg,offset);
#endif
    if(ptr)
    {
        rect r = { src_x,src_y, src_width,src_height };
        sender.Blit(/*width,height,bpp,data,bytes_per_line*/
                    total_width,total_height,
                    32,//depth,
                    (char*)ptr,
                    total_width*4,//depth/8,
                    r);
    }
    return res;
}

#define shmat_f_params int shmid,const void* shmaddr,int shmflag
void *shmat(shmat_f_params)
{
    static struct caller
    {
        void* (*orig) (shmat_f_params);
    public:
        caller()
        {
            void *f = dlopen("libXext.so", RTLD_NOW);
            if(!f) perror("libXext.so");
            orig = (void*(*)(shmat_f_params))dlsym(f, "shmat");
            if(!orig) perror("dlsym");
            dlclose(f);
        }
    } caller;

    void* res = caller.orig(shmid,shmaddr,shmflag);
    shmid_mappings[shmid] = res;
    return res;
}

#define xcb_shm_attach_f_params xcb_connection_t*c,xcb_shm_seg_t shmseg,uint32_t shmid,uint8_t readonly
xcb_void_cookie_t xcb_shm_attach(xcb_shm_attach_f_params)
{
    fprintf(stderr, "xcb_shm_attach detected\n");
    static struct caller
    {
        xcb_void_cookie_t (*orig) (xcb_shm_attach_f_params);
    public:
        caller()
        {
            void *f = dlopen("libxcb-shm.so", RTLD_NOW);
            if(!f) perror("libxcb-shm.so");
            orig = (xcb_void_cookie_t(*)(xcb_shm_attach_f_params))dlsym(f, "xcb_shm_attach");
            if(!orig) perror("dlsym");
            dlclose(f);
        }
    } caller;

    shmseg_mappings[shmseg] = shmid;

    return caller.orig(c,shmseg,shmid,readonly);
}

int ioctl(int fd,unsigned req,void *p)
{
    /* The old function is found in libc.so.6 */
    static struct Get1
    {
        void *f;
        Get1() : f(dlopen("libc.so.6",RTLD_NOW)) {}
        ~Get1() { dlclose(f); }
        void *operator() (const char *s) { return dlsym(f, s); }
    } caller;

    /* Find the real ioctl function */
    typedef int (*old_t)(int,unsigned,void *);
    static old_t old = NULL;
    if(!old) old = (old_t)caller("ioctl");

    /* Check this is an ALSA function call */
    //fprintf(stderr, "IOCTL type: 0x%08X\n", req);
    if(req == 0x400C4150  /* SNDRV_PCM_IOCTL_WRITEI_FRAMES */)
    {
        int result = old(fd, req, p);

        struct snd_xferi_t {
            long result;
            void *buf;
            unsigned long frames;
        }* xferi = (snd_xferi_t*) p;

        if(xferi->result > 0)
        {
            //fprintf(stderr, "Logging audio %p, res=%ld,frames=%lu\n", xferi->buf, xferi->result, xferi->frames);
            GetSender().AddAudio( (unsigned char*) xferi->buf, xferi->result*8);
        }

        return result;
    }
    if(req == 0xC004500A) // SNDCTL_DSP_SETFRAGMENT
    {
        dsp_filehandle = fd;
    }
    if(req == 0xC0844123 || req == 0xC0848123)
    {
        //fprintf(stderr, "Sound synced. Now what?\n");
    }
    if(req == 0xC25C4110) // SNDRV_PCM_IOCTL_HW_REFINE
    {
        struct snd_mask {
            uint_least32_t bits[(256+31)/32];
        };
        struct snd_interval {
            unsigned int min, max;
            unsigned int openmin:1,
                     openmax:1,
                     integer:1,
                     empty:1;   
        };
        struct snd_pcm_hw_params {
            unsigned int flags;
            struct snd_mask masks[2-0+1];
            struct snd_mask mres[5];    /* reserved masks */  
            struct snd_interval intervals[19-8+1];
            struct snd_interval ires[9];    /* reserved intervals */
            unsigned int rmask;     /* W: requested masks */
            unsigned int cmask;     /* R: changed masks */
            unsigned int info;      /* R: Info flags for returned setup */
            unsigned int msbits;        /* R: used most significant bits */
            unsigned int rate_num;      /* R: rate numerator */  
            unsigned int rate_den;      /* R: rate denominator */
            unsigned long fifo_size;    /* R: chip FIFO size in frames */
            unsigned char reserved[64]; /* reserved for future */
        };
        snd_pcm_hw_params* params = (snd_pcm_hw_params*)p;
        int had_rmask = params->rmask;
        fprintf(stderr, "SNDRV_PCM_IOCTL_HW_REFINE: rmask=%08X, cmask=%08X, masks[0]=%08X,%08X,%08X,%08X - ", params->rmask, params->cmask,
            params->masks[0].bits[0],
            params->masks[0].bits[1],
            params->masks[0].bits[2],
            params->masks[0].bits[3]);
        int result = old(fd, req, p);
        if(had_rmask & 1)
        {
            params->masks[0].bits[0] &= ~7; // Disable all mmap types, leaving only readi/writei/readn/writen
        }
        fprintf(stderr, "after call(res=%d): rmask=%08X, cmask=%08X, masks[0]=%08X,%08X,%08X,%08X\n", result, params->rmask, params->cmask,
            params->masks[0].bits[0],
            params->masks[0].bits[1],
            params->masks[0].bits[2],
            params->masks[0].bits[3]);
        return result;
    }

/*
#include <stddef.h>
#include <sys/time.h>
#include <asm/ioctl.h>
#include <sound/asound.h>
int a = <ioctl_name>;
gcc -E tmp.c > tmp2.c;clang -m32 -S -o- tmp2.c

00004112 : SNDRV_PCM_IOCTL_HW_FREE
00004122 : SNDRV_PCM_IOCTL_HWSYNC
00004140 : SNDRV_PCM_IOCTL_PREPARE      used by PapersPlease
00004141 : SNDRV_PCM_IOCTL_RESET
00004142 : SNDRV_PCM_IOCTL_START        used by PapersPlease
00004143 : SNDRV_PCM_IOCTL_DROP
00004144 : SNDRV_PCM_IOCTL_DRAIN
00004147 : SNDRV_PCM_IOCTL_RESUME
00004148 : SNDRV_PCM_IOCTL_XRUN
00004161 : SNDRV_PCM_IOCTL_UNLINK
000054A0 : SNDRV_TIMER_IOCTL_START
000054A1 : SNDRV_TIMER_IOCTL_STOP
000054A2 : SNDRV_TIMER_IOCTL_CONTINUE
000054A3 : SNDRV_TIMER_IOCTL_PAUSE
40044102 : SNDRV_PCM_IOCTL_TSTAMP
40044103 : SNDRV_PCM_IOCTL_TTSTAMP
40044145 : SNDRV_PCM_IOCTL_PAUSE
40044146 : SNDRV_PCM_IOCTL_REWIND
40044149 : SNDRV_PCM_IOCTL_FORWARD
40044160 : SNDRV_PCM_IOCTL_LINK
40045402 : SNDRV_TIMER_IOCTL_TREAD
40045532 : SNDRV_CTL_IOCTL_PCM_PREFER_SUBDEVICE         used by PapersPlease
40045542 : SNDRV_CTL_IOCTL_RAWMIDI_PREFER_SUBDEVICE
400C4150 : SNDRV_PCM_IOCTL_WRITEI_FRAMES
400C4152 : SNDRV_PCM_IOCTL_WRITEN_FRAMES
40345410 : SNDRV_TIMER_IOCTL_SELECT
403C5404 : SNDRV_TIMER_IOCTL_GPARAMS
40405514 : SNDRV_CTL_IOCTL_ELEM_LOCK
40405515 : SNDRV_CTL_IOCTL_ELEM_UNLOCK
40505412 : SNDRV_TIMER_IOCTL_PARAMS
80044100 : SNDRV_PCM_IOCTL_PVERSION
80044121 : SNDRV_PCM_IOCTL_DELAY
80045400 : SNDRV_TIMER_IOCTL_PVERSION
80045500 : SNDRV_CTL_IOCTL_PVERSION
80045530 : SNDRV_CTL_IOCTL_PCM_NEXT_DEVICE
800455D1 : SNDRV_CTL_IOCTL_POWER_STATE
800C4151 : SNDRV_PCM_IOCTL_READI_FRAMES
800C4153 : SNDRV_PCM_IOCTL_READN_FRAMES
80104132 : SNDRV_PCM_IOCTL_CHANNEL_INFO
80585414 : SNDRV_TIMER_IOCTL_STATUS
806C4120 : SNDRV_PCM_IOCTL_STATUS
80DC5521 : SNDRV_CTL_IOCTL_HWDEP_INFO
80E05411 : SNDRV_TIMER_IOCTL_INFO
81204101 : SNDRV_PCM_IOCTL_INFO
81785501 : SNDRV_CTL_IOCTL_CARD_INFO
C0045516 : SNDRV_CTL_IOCTL_SUBSCRIBE_EVENTS
C0045520 : SNDRV_CTL_IOCTL_HWDEP_NEXT_DEVICE
C0045540 : SNDRV_CTL_IOCTL_RAWMIDI_NEXT_DEVICE
C00455D0 : SNDRV_CTL_IOCTL_POWER
C008551A : SNDRV_CTL_IOCTL_TLV_READ
C008551B : SNDRV_CTL_IOCTL_TLV_WRITE
C008551C : SNDRV_CTL_IOCTL_TLV_COMMAND
C0145401 : SNDRV_TIMER_IOCTL_NEXT_DEVICE
C0405405 : SNDRV_TIMER_IOCTL_GSTATUS
C0405519 : SNDRV_CTL_IOCTL_ELEM_REMOVE
C0485510 : SNDRV_CTL_IOCTL_ELEM_LIST
C0684113 : SNDRV_PCM_IOCTL_SW_PARAMS        used by PapersPlease
C0844123 : SNDRV_PCM_IOCTL_SYNC_PTR         used by PapersPlease
C0E05403 : SNDRV_TIMER_IOCTL_GINFO
C10C5541 : SNDRV_CTL_IOCTL_RAWMIDI_INFO
C1105511 : SNDRV_CTL_IOCTL_ELEM_INFO
C1105517 : SNDRV_CTL_IOCTL_ELEM_ADD
C1105518 : SNDRV_CTL_IOCTL_ELEM_REPLACE
C1205531 : SNDRV_CTL_IOCTL_PCM_INFO
C25C4110 : SNDRV_PCM_IOCTL_HW_REFINE        used by PapersPlease
C25C4111 : SNDRV_PCM_IOCTL_HW_PARAMS        used by PapersPlease
C2C45512 : SNDRV_CTL_IOCTL_ELEM_READ
C2C45513 : SNDRV_CTL_IOCTL_ELEM_WRITE

*/

    /* Call the real ioctl function */
    return old(fd,req,p);
}

#if 0
ssize_t write(int fd,const void* buf, std::size_t nbytes)
{
    static struct caller
    {
        ssize_t (*orig)(int,const void*,std::size_t);
    public:
        caller()
        {
            void *f = dlopen("libc.so.6", RTLD_NOW);
            if(!f) perror("libc.so.6");
            orig = (decltype(orig))dlsym(f, "write");
            if(!orig) perror("dlsym");
            dlclose(f);
        }
    } caller;
    if(fd > 0 && fd == dsp_filehandle)
    {
        fprintf(stderr, "DSP write %u bytes\n", (int)nbytes);
        fflush(stderr);
    }
    else
    {
        fprintf(stderr, "FD %d write %u bytes (dsp=%d)\n", fd,(int)nbytes,dsp_filehandle);
        fflush(stderr);
    }
    return caller.orig(fd,buf,nbytes);
}

int snd_pcm_hw_params_set_access(void* pcm, void* params, int access)
{
    /* The old function is found in libasound.so */
    static struct Get2
    {
        void *f;
        Get2() : f(dlopen("libasound.so", RTLD_LAZY)) {}
        ~Get2() { dlclose(f); }
        void *operator() (const char *s) { return dlsym(f, s); }
    } caller;

    /* mmap is impossible to hook, thus we sort of disable it! */
    if(access == 0 /* SND_PCM_ACCESS_MMAP_INTERLEAVED */) return -1;

    /* Find the real function */
    typedef int (*old_t)(void*, void*, int);
    static old_t old = NULL;
    if(!old) old = (old_t)caller("snd_pcm_hw_params_set_access");

    fprintf(stderr, "snd_pcm_hw_params_set_access called\n");

    /* Call the real function */
    return old(pcm,params,access);
}
#endif

}
